using System;
using System.Math;
using System.Threading;
using System.Diagnostics;


// Empty Benchmark for comparison
// using async methods up and down is very slow, but making them void is too fast

// virtual methods are less likely to be inlined
public abstract class AbstractFSM {
  public abstract void down(long n);
  public abstract void up(long n);
  public abstract long block();
}

class FSM : AbstractFSM {
  private long level;
  private async unblock_u();
  private async unblock_d();
  double result1;
  double result2;
  override public void down(long n) {
    down_inner(n) ;
  }
  public void down_inner(long n) {
    if (n>=level) {
      unblock_d();
      try {// exception code lower inlining probability
	if (n/(n-level)==0) 
	  unblock_d();
      } catch (System.DivideByZeroException dbz) {
	unblock_d();
      }
    }
    result1=Math.Sin(n*1.001);
  }   
  override public void up(long n) {
   up_inner(n);
  }  
  private void up_inner(long n) {
   if (n>=level) {
      unblock_u();
      try { // exception code lower inlining probability
	if (n/(n-level)==0) 
	  unblock_u();
      } catch (System.DivideByZeroException dbz) {
	unblock_u();
      }
    }
    result2=Math.Sin(n*1.002);
  }
  override public long block() { // can not be part of chord
    return innerBlock();
  } 
  private long innerBlock() & unblock_u() & unblock_d() {
    return DateTime.Now.Ticks;
  }

  public FSM(long _level) {
    level=_level;

  }
}

public class Program {
  AbstractFSM fsm;
  long count;
  public Program(long _count) {
    //Console.WriteLine("Constructor");
    this.count=_count;
    fsm=new FSM(_count);
    //Console.WriteLine("constructor done");
  }
  public static void Main() {
    for (int i=0;i<=5;i++) {
      Program p=new Program(500000);
      p.start_measurement();
      Thread.Sleep(4000);
    }
  }
  private async upThread() {
    for (long iter=0;iter<=count;iter++) {
      fsm.up(iter);
    }
  }
  private async downThread() {
    for (long iter=0;iter<=count;iter++) {
      fsm.down(iter);
    } 
  }
  public void start_measurement() {
    long before=DateTime.Now.Ticks;
    upThread();
    downThread();
    long after=fsm.block(); // wait until both event sequesnces finish
    double ms=new TimeSpan(after-before).TotalMilliseconds;
    Console.WriteLine(""+ms+"   normalize 1k "+(1000*ms/count)+" , "+count+" cycles")   ;
  }

}
